import pandas as pd
import numpy as np

import astral
import datetime
import pvlib


class filter_single:
    """
    Class to execute filters for individual PV systems
    """

    def min_availability(df_pv,lat,lon,tz,share):
        """
        Function description
        --------------------
        Filter removes days from dataset if the data availabily rate of
        power measurements for zenith < 85 is below share
        
        df_pv should be pandas dataframe with index datetime values and first 
        column power measurements

        Function parameters
        -------------------
        df_pv : pandas dataframe
            PV power measurments.
        lat : numeric
            Latitude.
        lon : numeric
            Longitude.
        tz : string
            Timezone.
        share : numeric
            Threshold share of minimum daily data availability rate.

        Returns
        -------
        df_out : pandas dataframe
            Filtered PV power measurements.

        """
 
        col = df_pv.columns[0]
        
        df_pv.loc[:,'date'] = df_pv.index.date
        
        remove_days = []
        
        for day in df_pv['date'].unique():
            
            df_temp = pd.DataFrame()
            df_temp = df_pv[df_pv['date']==day].copy(deep=True)
            df_temp.loc[:,'zenith'] = pvlib.location.Location(lat,lon, tz=tz).get_solarposition(df_temp.index)['zenith']
            df_temp = df_temp.loc[df_temp['zenith']<85]
            
            if df_temp.count()[0]<int(np.round(share*len(df_temp))): 
                remove_days = remove_days +[day]
                
        df_out = df_pv.loc[df_pv['date'].isin(remove_days),col] = np.nan
        
        return df_out


    def night_time(df_pv,df_w,lat,lon,alt,tz,tilt,azimuth,dc_capacity,coef_temp,coef_inv,coef_inv_norm,model_par,ct,cc):
        """
        Function description
        --------------------
        Filter sets night time values to zero by simulating the PV power output 
        for clear sky conditions, considering the PV system characteristics
        
        Power output values are set to zero if the simulated clear sky power output 
        is zero

        Function parameters
        -------------------
        df_pv : pandas dataframe
            PV power measurments.
        df_w : pandas dataframe
            Recording of weather variables.
        lat : numeric
            Latitude.
        lon : numeric
            Longitude.
        tz : string
            Timezone.
        alt : numeric
            Altitude.
        tilt : numeric
            Tilt angle.
        azimuth : numeric
            Azimuth angle.
        dc_capacity : numeric
            Rated DC capacity.
        coef_temp : numeric
            The temperature coefficient in units of 1/C.
        coef_inv : numeric
            Nominal inverter efficiency.
        coef_inv_norm : numeric
            Reference inverter efficiency.
        model_par : dictionary
            Temperature model parameters.
        ct : string
            City.
        cc : string
            Country code.

        Returns
        -------
        df_pv : pandas dataframe
            Filtered PV power measurements.

        """

        col = df_pv.columns[0]
        
        df_temp = pd.merge(df_pv, df_w, left_index=True, right_index = True, how='left')
    
        df_temp['extra'] = pvlib.irradiance.get_extra_radiation(df_temp.index)  
        
        df_az = pvlib.location.Location(lat,lon, tz=tz, name=ct+'/'+cc).get_solarposition(df_temp.index)  
        
        df_simplified_solis = pvlib.location.Location(lat, lon, tz=tz).get_clearsky(df_temp.index, model = 'simplified_solis')
       
        rel_arimass = pvlib.atmosphere.get_relative_airmass(df_az['apparent_zenith'], model='kastenyoung1989') 

        df_poa = pvlib.irradiance.get_total_irradiance(tilt, azimuth, df_az['zenith'], df_az['azimuth'], 
                                                       df_simplified_solis['dni'],df_simplified_solis['ghi'], 
                                                       df_simplified_solis['dhi'],dni_extra=df_temp['extra'],
                                                       airmass=rel_arimass, albedo=0.25, surface_type='urban', 
                                                       model='perez')
            
        aoi = pvlib.irradiance.aoi(tilt, azimuth, df_az['zenith'], df_az['azimuth'])
    
        iam = pvlib.iam.physical(aoi) 
        
        cell_temp = pvlib.temperature.sapm_cell(df_poa['poa_global'], df_temp['amb_temp_C'], df_temp['wind_m_p_s'], 
                                               model_par['a'], model_par['b'], model_par['deltaT'], irrad_ref=1000.0)
             
        effective_irradiance = (df_poa['poa_direct'] * iam + df_poa['poa_diffuse'])
    
        pdc0 = 180 
        
        pdc = pvlib.pvsystem.pvwatts_dc(effective_irradiance, cell_temp, pdc0, coef_temp)
        pac = pvlib.inverter.pvwatts(pdc, pdc0,eta_inv_nom=coef_inv, eta_inv_ref = coef_inv_norm)
        pnac = pac/pdc0*dc_capacity
    
        df_pv['clear sky production'] = pnac.fillna(0)
    
        df_pv.loc[df_pv.index[df_pv['clear sky production']==0],col]=0
    
        df_pv = df_pv[[col]]
    
        return df_pv


    def around_sundown(df_pv,lat,lon,alt,tz,ct,cc):
        """
        Function description
        --------------------
        Filter sets negative values that occur during sunrise and sunset to zero.

        Function parameters
        -------------------
        df_pv : pandas dataframe
            PV power measurments.
        lat : numeric
            Latitude.
        lon : numeric
            Longitude.
        alt : numeric
            Altitude.
        tz : string
            Timezone.
        ct : string
            City.
        cc : string
            Country code.

        Returns
        -------
        df_out : pandas dataframe
            Filtered PV power measurements.

        """

        col = df_pv.columns[0]
    
        location = astral.Location((ct, cc, lat, lon, tz, alt))

        df_pv.loc[:,'date'] = df_pv.index.date
        df_pv.loc[:,'sun_up'] = 1 
        
        df_out = pd.DataFrame()

        for day in df_pv['date'].unique():
            df_temp = pd.DataFrame()
            df_temp = df_pv[df_pv['date']==day].copy(deep=True)

            sun = location.sun(date=day, local=True)
            sunrise, sunset = sun['sunrise']+ datetime.timedelta(hours=.5), sun['sunset']+ datetime.timedelta(hours=-.5)
            
            df_temp.loc[df_temp.index<sunrise,'sun_up']=0
            df_temp.loc[df_temp.index>sunset,'sun_up']=0
            
            if df_out.empty:
                df_out = df_temp
            else:
                df_out = df_out.append(df_temp)

        df_out.loc[(df_out[col]<0 )&(df_out['sun_up']==0),col] = 0
        df_out = df_out[[col]]
     
        return df_out
    
    

    def lower_limit(df_pv):
        """
        Function description
        --------------------
        Filter sets negative values to NaN

        Function parameters
        -------------------
        df_pv : pandas dataframe
            PV power measurments.

        Returns
        -------
        df_pv : pandas dataframe
            Filtered PV power measurements.

        """
        col = df_pv.columns[0]
        df_pv.loc[df_pv[col]<0,col]=np.nan
        
        return df_pv
        
    
    def upper_limit(df_pv,df_w,lat,lon,alt,tz,tilt,azimuth,dc_capacity,ac_capacity,coef_temp,coef_inv,coef_inv_norm,model_par,ct,cc,kpv):
        """
        Function description
        --------------------
        Filter sets values that exceed the maximum expected production of a PV system 
        for clear sky conditions to NaN

        Function parameters
        -------------------
        df_pv : pandas dataframe
            PV power measurments.
        df_w : pandas dataframe
            Recording of weather variables.
        lat : numeric
            Latitude.
        lon : numeric
            Longitude.
        alt : numeric
            Altitude.
        tz : string
            Timezone.
        tilt : numeric
            Tilt angle.
        azimuth : numeric
            Azimuth angle.
        dc_capacity : numeric
            Rated DC capacity.
        ac_capacity : numeric
            Rated AC capacity.
        coef_temp : numeric
            The temperature coefficient in units of 1/C.
        coef_inv : numeric
            Nominal inverter efficiency.
        coef_inv_norm : numeric
            Reference inverter efficiency.
        model_par : dictionary
            Temperature model parameters.
        ct : string
            City.
        cc : string
            Country code.
        kpv : numeric
            Threshold value of allowed cloud enhancement effect.

        Returns
        -------
        df_temp : pandas dataframe
            Filtered PV power measurements.

        """

        col = df_pv.columns[0]
        
        df_temp = pd.merge(df_pv, df_w, left_index=True, right_index = True, how='left')
        
        # AC capacity component
        df_temp['Power'][df_temp['Power']>ac_capacity*1.025] = np.nan
    
        # DC capacity component
        df_temp['extra'] = pvlib.irradiance.get_extra_radiation(df_temp.index)  
        
        df_az = pvlib.location.Location(lat,lon, tz=tz, name=ct+'/'+cc).get_solarposition(df_temp.index)  
        
        df_simplified_solis = pvlib.location.Location(lat, lon, tz=tz).get_clearsky(df_temp.index, model = 'simplified_solis')
        
        rel_arimass = pvlib.atmosphere.get_relative_airmass(df_az['apparent_zenith'], model='kastenyoung1989') 
    
        df_poa = pvlib.irradiance.get_total_irradiance(tilt, azimuth, df_az['zenith'], df_az['azimuth'], 
                                                       df_simplified_solis['dni'],df_simplified_solis['ghi'], 
                                                       df_simplified_solis['dhi'],dni_extra=df_temp['extra'],
                                                       airmass=rel_arimass, albedo=0.25, surface_type='urban', 
                                                       model='perez')
    
        aoi = pvlib.irradiance.aoi(tilt, azimuth, df_az['zenith'], df_az['azimuth'])
    
        iam = pvlib.iam.physical(aoi) 
        cell_temp = pvlib.temperature.sapm_cell(df_poa['poa_global'], df_temp['amb_temp_C'], df_temp['wind_m_p_s'], 
                                               model_par['a'], model_par['b'], 
                                               model_par['deltaT'], irrad_ref=1000.0)
             
        effective_irradiance = (df_poa['poa_direct'] * iam + df_poa['poa_diffuse'])
    
        pdc0=180
        
        pdc = pvlib.pvsystem.pvwatts_dc(effective_irradiance, cell_temp, pdc0, coef_temp)
        pac = pvlib.inverter.pvwatts(pdc, pdc0, eta_inv_nom=coef_inv, eta_inv_ref = coef_inv_norm)
        
        pnac = pac/pdc0
        
        pv_ac = pnac*dc_capacity
        
        df_temp.loc[:,'upper_limit'] = pv_ac*kpv
        
        df_temp.loc[:,'zenith'] = df_az['zenith']
        
        df_temp.loc[:,'Power2'] = df_temp[col]
        df_temp.loc[(df_temp['Power2'] > df_temp['upper_limit']) & (df_temp['zenith']<80),'Power2'] = np.nan
        df_temp.loc[(df_temp['Power2'] > df_temp['upper_limit'])
                              &(df_temp['zenith']>=80)
                              &(df_temp['zenith']<85)
                              &(df_temp['Power2']>.125*dc_capacity), 'Power2'] = np.nan    
        df_temp.loc[(df_temp['Power2'] > df_temp['upper_limit'])
                          &(df_temp['zenith']>=85)
                          &(df_temp['Power2']>.075*dc_capacity), 'Power2'] = np.nan        
    
        df_temp = df_temp[['Power2']]
        df_temp = df_temp.rename(columns={'Power2': col})
    
        return df_temp



    def consecutive_period(df_pv,lat,lon,tz, max_consec):
        """
        Function description
        --------------------
        Filter sets values to NaN in case the power measurements feature a constant or 
        linear increasing/decreasing value for a consecutive period equal to or greater 
        than max_consec      

        Function parameters
        -------------------
        df_pv : pandas dataframe
            PV power measurments.
        lat : numeric
            Latitude.
        lon : numeric
            Longitude.
        tz : string
            Timezone.
        max_consec : numeric
            Threshold value setting the maximum allowed sequence of constant or linear 
            increasing/decreasing power measurements.

        Returns
        -------
        df_out : pandas dataframe
            Filtered PV power measurements.

        """

        col = df_pv.columns[0]
        
        df_pv.loc[:,'date'] = df_pv.index.date
        
        for day in df_pv['date'].unique():
            df_temp = pd.DataFrame()

            df_temp = df_pv[df_pv['date']==day].copy(deep=True)
            df_temp.loc[:,'zenith'] =  pvlib.location.Location(lat,lon, tz=tz).get_solarposition(df_temp.index)['zenith']
            df_temp = df_temp.loc[df_temp['zenith']<85]
            df_temp.loc[:,'diff'] = df_temp[col].diff()
            df_temp.loc[:,'flag'] = df_temp['diff'].groupby([df_temp['diff'].diff().ne(0).cumsum()]).transform('size').ge(max_consec).astype(int)
            df_temp = df_temp.loc[df_temp['flag'] == 1]
            
            if df_temp.empty == False:
                df_pv.loc[df_temp.index, col] = np.nan
                
        df_out =  df_pv[[col]]  
        
        return df_out
    

    
    def persistence(df_pv,lat,lon,tz,dc_capacity):
        """
        Function description
        --------------------
        Filter removes entire days from dataset if the power output variability for 
        zenith < 80 is extremely low or high

        Function parameters
        -------------------
        df_pv : pandas dataframe
            PV power measurments.
        lat : numeric
            Latitude.
        lon : numeric
            Longitude.
        tz : string
            Timezone.
        dc_capacity : numeric
            Rated DC capacity.

        Returns
        -------
        df_pv : pandas dataframe
            Filtered PV power measurements.

        """
            
        col = df_pv.columns[0]
        
        df_pv.loc[:,'date'] = df_pv.index.date
        
        remove_days = []
        
        for day in df_pv['date'].unique():
            
            df_temp = pd.DataFrame()
            
            df_temp = df_pv[df_pv['date']==day].copy(deep=True)
            df_temp.loc[:,'zenith'] = pvlib.location.Location(lat,lon, tz=tz).get_solarposition(df_temp.index)['zenith']
            
            df_temp = df_temp.loc[df_temp['zenith']<80]

            df_temp.loc[:,'extra'] = pvlib.irradiance.get_extra_radiation(df_temp.index)
            df_temp.loc[:,'extra_norm'] =  np.cos(np.radians(df_temp['zenith'])) * df_temp['extra'] / 1000
            df_temp.loc[:,'Power_norm'] = df_temp[col]/dc_capacity
            lower_b = 1/8* ((df_temp[col]/dc_capacity)/(df_temp['extra_norm'])).mean()
            upper_b = 0.35
            test_value = ((df_temp[col]/dc_capacity)/(df_temp['extra_norm'])).std()
            
            if (lower_b > test_value):
                remove_days = remove_days +[day]
            if (test_value > upper_b):
                remove_days = remove_days +[day]        
        df_pv.loc[df_pv['date'].isin(remove_days),col] = np.nan
        return df_pv
        


class filter_across:
    
    def daily_energy_ratio(df_pv,df_metadata,df_w,lat,lon,tz,alt,coef_temp,coef_inv,coef_inv_norm,max_der_deviation=.25):
        """
        Function description
        --------------------
        Sets all values on a day to NaN in case the daily energy ratio deviates more than
        max_der_deviation obtained from the mean daily energy ratio of all systems

        Function parameters
        -------------------
        df_pv : pandas dataframe
            PV power measurments.
        df_metadata : pandas dataframe
            PV system information.
        df_w : pandas dataframe
            Recording of weather variables.
        lat : numeric
            Latitude.
        lon : numeric
            Longitude.
        tz : string
            Timezone.
        alt : numeric
            Altitude.
        coef_temp : numeric
            The temperature coefficient in units of 1/C.
        coef_inv : numeric
            Nominal inverter efficiency.
        coef_inv_norm : numeric
            Reference inverter efficiency.
        max_der_deviation : numeric, optional
            Threshold value for maximum deviation of daily energy ratio. The default is 0.25.

        Returns
        -------
        df_pv : pandas dataframe
            Filtered PV power measurements.

        """            
    
        df_der = support.construct_der_matrix(df_pv,df_metadata,df_w,lat,lon,tz,alt,coef_temp,coef_inv,coef_inv_norm)
    
        df_der_original = df_der.copy(deep=True)
        df_der_filter = df_der.copy(deep=True)   

        df_pv['date'] = df_pv.index.date

        for der_id in df_der_filter.columns[:-1]:
            system_id = der_id.split("_")[0]  
            df_der_filter[der_id]=df_der_filter[der_id][(df_der_filter[der_id]< df_der_filter['mean']+max_der_deviation) & 
                                                        (df_der_filter[der_id]>df_der_filter['mean']-max_der_deviation)]
            df_der_filter_single = df_der_filter[[der_id]]
            df_der_original_single = df_der_original[[der_id]]
            dif = pd.concat([df_der_original_single, df_der_filter_single]).drop_duplicates(keep=False).index.date
            df_pv.loc[df_pv['date'].isin(dif),system_id] = np.nan
            
        df_pv = df_pv.drop('date',axis=1)
            
        return df_pv




class support:
    
    def construct_der_matrix(df_pv,df_metadata,df_w,lat,lon,tz,alt,coef_temp,coef_inv,coef_inv_norm,normalize_csi=True):
        """
        Function description
        --------------------
        Constructs matrix of daily energy ratio

        Function parameters
        -------------------
        df_pv : pandas dataframe
            PV power measurments.
        df_metadata : pandas dataframe
            PV system information.
        df_w : pandas dataframe
            Recording of weather variables.
        lat : numeric
            Latitude.
        lon : numeric
            Longitude.
        tz : string
            Timezone.
        alt : numeric
            Altitude.
        coef_temp : numeric
            The temperature coefficient in units of 1/C.
        coef_inv : numeric
            Nominal inverter efficiency.
        coef_inv_norm : numeric
            Reference inverter efficiency.
        normalize_csi : boolean, optional
            Normalize PV power output values to expected power output during clear sky conditions, 
            to be applied in case PV systems have different rated capacities. The default is True.

        Returns
        -------
        df_der : dataframe
            Daily energy ration per system per day.

        """
        df_der = pd.DataFrame()
        for system_id in df_pv.columns:
            tilt = df_metadata.loc[df_metadata['ID'] == system_id]['tilt'].item()
            azimuth = df_metadata.loc[df_metadata['ID'] == system_id]['azimuth'].item()
            dc_capacity = df_metadata.loc[df_metadata['ID'] == system_id]['estimated_dc_capacity'].item()

            df_temp = pd.DataFrame()        
            df_temp = df_pv[[system_id]]
            df_temp = df_temp.dropna()
            df_pv_csi = support.csi_pv(df_temp,df_w,lat,lon,tz,alt,tilt,azimuth,dc_capacity,system_id,coef_temp,coef_inv,coef_inv_norm)
        
            df_pv_csi_day = df_pv_csi.resample('D').sum()
            df_pv_csi_day[system_id+'_der'] = df_pv_csi_day[system_id]/df_pv_csi_day[system_id+'_csi']
            df_pv_csi_day =df_pv_csi_day[[system_id+'_der']]
            if df_der.empty:
                df_der = df_pv_csi_day
            else:
                df_der = pd.merge(df_der,df_pv_csi_day,left_index=True,right_index=True,how='outer')

        if normalize_csi ==True:
            for der_id in df_der.columns:
                df_der[der_id] = df_der[der_id]/df_der[der_id].max()

        df_der['mean'] = df_der.mean(axis=1)
        
        return df_der
        
    
    
    
    def csi_pv(df_pv,df_w,lat,lon,tz,alt,tilt,azimuth,dc_capacity,system_id,coef_temp,coef_inv,coef_inv_norm):
        """
        Function description
        --------------------
        Estimates the PV power output for clear sky conditions

        Function parameters
        -------------------
        df_pv : pandas dataframe
            PV power measurments.
        df_w : pandas dataframe
            Recording of weather variables.
        lat : numeric
            Latitude.
        lon : numeric
            Longitude.
        tz : string
            Timezone.
        alt : numeric
            Altitude.
        tilt : numeric
            Tilt angle.
        azimuth : numeric
            Azimuth angle.
        dc_capacity : numeric
            Rated DC capacity.
        system_id : numeric
            PV system ID, column name.
        coef_temp : numeric
            The temperature coefficient in units of 1/C.
        coef_inv : numeric
            Nominal inverter efficiency.
        coef_inv_norm : numeric
            Reference inverter efficiency.


        Returns
        -------
        df_pv : pandas dataframe
            PV power output for clear sky conditions.

        """
        
        
        df_temp = pd.merge(df_pv, df_w, left_index=True, right_index = True, how='left')
    
        df_temp['extra'] = pvlib.irradiance.get_extra_radiation(df_temp.index)  
        
        df_az = pvlib.location.Location(lat,lon, tz=tz).get_solarposition(df_temp.index)  
        
        df_simplified_solis = pvlib.location.Location(lat, lon, tz=tz).get_clearsky(df_temp.index, model = 'simplified_solis')
       
        rel_arimass = pvlib.atmosphere.get_relative_airmass(df_az['apparent_zenith'], model='kastenyoung1989')

        df_poa = pvlib.irradiance.get_total_irradiance(tilt, azimuth, df_az['zenith'], df_az['azimuth'], 
                                                       df_simplified_solis['dni'],df_simplified_solis['ghi'], 
                                                       df_simplified_solis['dhi'],dni_extra=df_temp['extra'],
                                                       airmass=rel_arimass, albedo=0.25, surface_type='urban', 
                                                       model='perez')
            
        aoi = pvlib.irradiance.aoi(tilt, azimuth, df_az['zenith'], df_az['azimuth'])
    
        iam = pvlib.iam.physical(aoi)
        
        temperature_model_parameters ={'a': -2.98, 'b': -0.0471, 'deltaT': 1}
        cell_temp = pvlib.temperature.sapm_cell(df_poa['poa_global'], df_temp['amb_temp_C'], df_temp['wind_m_p_s'], 
                                               temperature_model_parameters['a'], temperature_model_parameters['b'], 
                                               temperature_model_parameters['deltaT'], irrad_ref=1000.0)
             
        effective_irradiance = (df_poa['poa_direct'] * iam + df_poa['poa_diffuse'])
    
        pdc0 = 180 
        pdc = pvlib.pvsystem.pvwatts_dc(effective_irradiance, cell_temp, pdc0, coef_temp)
        pac = pvlib.inverter.pvwatts(pdc, pdc0,eta_inv_nom=coef_inv, eta_inv_ref = coef_inv_norm)
    
        pnac = pac/pdc0*dc_capacity
    
        df_pv[system_id+'_csi'] = pnac.fillna(0)
    
        df_pv = df_pv[[system_id, system_id+'_csi']]
        return df_pv
    
    
    
    
    